// +build js

package mp3

import (
	"errors"
	"fmt"
	"io"
	"io/ioutil"
	"math"
	"time"

	"engo.io/engo/common/internal/decode/convert"

	"github.com/gopherjs/gopherjs/js"
)

// TODO: This just uses decodeAudioData, that can treat audio files other than MP3.

type Stream struct {
	leftData   []float32
	rightData  []float32
	posInBytes int
}

var (
	errTryAgain = errors.New("audio/mp3: try again")
	errTimeout  = errors.New("audio/mp3: timeout")
)

func (s *Stream) Read(b []byte) (int, error) {
	l := len(s.leftData)*4 - s.posInBytes
	if l > len(b) {
		l = len(b)
	}
	l = l / 4 * 4
	const max = 1<<15 - 1
	for i := 0; i < l/4; i++ {
		il := int32(s.leftData[s.posInBytes/4+i] * max)
		if il > max {
			il = max
		}
		if il < -max {
			il = -max
		}
		ir := int32(s.rightData[s.posInBytes/4+i] * max)
		if ir > max {
			ir = max
		}
		if ir < -max {
			ir = -max
		}
		b[4*i] = byte(il)
		b[4*i+1] = byte(il >> 8)
		b[4*i+2] = byte(ir)
		b[4*i+3] = byte(ir >> 8)
	}
	s.posInBytes += l
	if s.posInBytes == len(s.leftData)*4 {
		return l, io.EOF
	}
	return l, nil
}

func (s *Stream) Seek(offset int64, whence int) (int64, error) {
	next := int64(0)
	switch whence {
	case io.SeekStart:
		next = offset
	case io.SeekCurrent:
		next = int64(s.posInBytes) + offset
	case io.SeekEnd:
		next = s.Length() + offset
	}
	s.posInBytes = int(next)
	return next, nil
}

func (s *Stream) Close() error {
	return nil
}

func (s *Stream) Length() int64 {
	return int64(len(s.leftData) * 4)
}

// Size is deprecated as of 1.6.0-alpha. Use Length instead.
func (s *Stream) Size() int64 {
	return s.Length()
}

// seekNextFrame seeks the next frame and returns the new buffer with the new position.
// seekNextFrame also returns true when seeking is successful, or false otherwise.
//
// Seeking is necessary when decoding fails. Safari's MP3 decoder can't treat IDs well (#438).
func seekNextFrame(buf []byte) ([]byte, bool) {
	// TODO: Need to skip tags explicitly? (hajimehoshi/go-mp3#9)

	if len(buf) < 1 {
		return nil, false
	}
	buf = buf[1:]

	for {
		if buf[0] == 0xff && buf[1]&0xfe == 0xfe {
			break
		}
		buf = buf[1:]
		if len(buf) < 2 {
			return nil, false
		}
	}
	return buf, true
}

func Decode(src convert.ReadSeekCloser, sr int) (*Stream, error) {
	b, err := ioutil.ReadAll(src)
	if err != nil {
		return nil, err
	}
	if err := src.Close(); err != nil {
		return nil, err
	}
	var s *Stream
	try := 0
	for {
		s, err = decode(sr, b, try)
		switch err {
		case errTryAgain:
			buf, ok := seekNextFrame(b)
			if !ok {
				return nil, fmt.Errorf("audio/mp3: Decode failed: invalid format?")
			}
			b = buf
			continue
		case errTimeout:
			try++
			continue
		default:
			if err != nil {
				return nil, err
			}
		}
		break
	}
	return s, nil
}

var offlineAudioContextClass *js.Object

func init() {
	if klass := js.Global.Get("OfflineAudioContext"); klass != js.Undefined {
		offlineAudioContextClass = klass
		return
	}
	if klass := js.Global.Get("webkitOfflineAudioContext"); klass != js.Undefined {
		offlineAudioContextClass = klass
		return
	}
}

func decode(sr int, buf []byte, try int) (*Stream, error) {
	if offlineAudioContextClass == nil {
		return nil, errors.New("audio/mp3: OfflineAudioContext is not available")
	}

	s := &Stream{}
	ch := make(chan error)

	// TODO: webkitOfflineAudioContext's constructor might causes 'Syntax' error
	// when the sample rate is not so usual like 22050 on Safari.
	// Should this be handled before calling this?

	// TODO: 1 is a correct second argument?
	oc := offlineAudioContextClass.New(2, 1, sr)
	oc.Call("decodeAudioData", js.NewArrayBuffer(buf), func(buf *js.Object) {
		s.leftData = buf.Call("getChannelData", 0).Interface().([]float32)
		switch n := buf.Get("numberOfChannels").Int(); n {
		case 1:
			s.rightData = s.leftData
			close(ch)
		case 2:
			s.rightData = buf.Call("getChannelData", 1).Interface().([]float32)
			close(ch)
		default:
			ch <- fmt.Errorf("audio/mp3: number of channels must be 1 or 2 but %d", n)
		}
	}, func(err *js.Object) {
		if err != nil {
			ch <- fmt.Errorf("audio/mp3: decodeAudioData failed: %v", err)
		} else {
			// On Safari, error value might be null and it is needed to retry decoding
			// from the next frame (#438).
			ch <- errTryAgain
		}
	})

	timeout := time.Duration(math.Pow(2, float64(try))) * time.Second

	select {
	case err := <-ch:
		if err != nil {
			return nil, err
		}
	case <-time.After(timeout):
		// Sometimes decode fails without calling the callbacks (#464).
		// Let's just try again in this case.
		return nil, errTimeout
	}
	return s, nil
}
